/*
* wmacpimon v0.2.1
* (C) 2002 Costantino Pistagna (valvoline@vrlteam.org)
* wmacpimon homepage: http://www.vlrteam.org/wmacpimon
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include "wmacpimon.h"
#include "acpiintf.h"
#include "parsers.h"

#define MAX(a,b)  ((a)>(b)?(a):(b))
#define MIN(a,b)  ((a)<(b)?(a):(b))

#ifdef ACPI
/* check if there's something at fName */
static int exists(char *fName) {
	struct stat fdata;

	return stat(fName, &fdata)==0;
}

void dump_acpinfo(ACPInfo *acpinfo) {
	BattInfo *batt=acpinfo->batteries;

	printf("Temperature (psv/crit): %d (%d/%d)\n",
		acpinfo->temperature, acpinfo->critical_temp,
		acpinfo->passive_temp);
	printf("CPU Speed: %d\n", acpinfo->speed);
	while (batt) {
		printf("Battery %s:\n", batt->statepath);
		printf("  Charge %d of %d, rate %d\n", batt->currcap, batt->fullcap,
			batt->currrate);
		printf("  Charging state: %d\n", batt->charge);
		batt = batt->next;
	}
	printf("Cumulative charge state: %d\n", acpinfo->cum_charge);
	printf("Cumulative percent: %d\n", acpinfo->battery_percent);
	printf("Cumulative remain/full: %d/%d\n", acpinfo->cum_remaining,
		acpinfo->cum_fullcap);
	printf("Remaining %d hrs, %d mins\n", acpinfo->remaining_hours,
		acpinfo->remaining_minutes);
}


/* see if we have ACPI support and check version */
int power_init(void) {
    FILE *acpi;
    char buf[4096];
    int acpi_ver = 0;
    if (!(acpi = fopen("/sys/module/acpi/parameters/acpica_version", "r"))) {
		fprintf(stderr, "This system does not support ACPI\n");
		return 1;
    }

    /* okay, now see if we got the right version */
    fread(buf, 4096, 1, acpi);
    acpi_ver = strtol(buf, NULL, 10);
    if (acpi_ver < 20020214) {
		fprintf(stderr, "This version requires ACPI subsystem version 20020214 or later\n");
		fclose(acpi);
		return 1;
    }
    /* yep, all good */
    fclose(acpi);

    return 0;
}


static int readIntegerFrom(char *basePath, char *fileName, int *dest) {
	char curPath[PATHLEN];
	FILE *src;
	int parsed;

	snprintf(curPath, PATHLEN, "%s/%s", basePath, fileName);
	if (!exists(curPath)) {
		return -1;
	}
	src = fopen(curPath, "r");
	parsed = fscanf(src, "%d", dest);
	fclose(src);
	return parsed;
}

static void parseChargeState(BattInfo *battinfo) {
	char curPath[PATHLEN];
	FILE *src;
	char chargeState[20];

	enum { CHARGING, DISCHARGING, FULL, INVALID };

	snprintf(curPath, PATHLEN, "%s/%s", battinfo->statepath, "status");
	if (!exists(curPath)) {
		battinfo->charge = INVALID;
		return;
	}
	src = fopen(curPath, "r");
	if (!fgets(chargeState, sizeof(chargeState), src))
		battinfo->charge =INVALID;
	if (!strcmp(chargeState, "Charging\n")) {
		battinfo->charge = CHARGING;
	} else if (!strcmp(chargeState, "Discharging\n")
			|| !strcmp(chargeState, "Not charging\n")) {
		battinfo->charge = DISCHARGING;
	} else if (!strcmp(chargeState, "Full\n")) {
		battinfo->charge = FULL;
	} else if (!strcmp(chargeState, "Unknown\n")) {
		/* "Unknown" occurs when currcap is greater then fullcap */
		if (battinfo->currcap >= battinfo->fullcap) {
			battinfo->charge = FULL;
		} else {
			battinfo->charge = INVALID;
		}
	}
	fclose(src);
}

static void update_battstate(BattInfo *battinfo) {
	/* get current capacity in uWh; return in mWh */
	if (!readIntegerFrom((char *)&battinfo->statepath, "energy_now", &battinfo->currcap))
		battinfo->currcap = 0;
	battinfo->currcap /= 1000;
	if (!readIntegerFrom((char *)&battinfo->statepath, "current_now",  &battinfo->currrate))
		battinfo->currrate = 0;
	battinfo->currrate /= 1000; /* give in milliwatt */
	/* get current charge state */
	parseChargeState(battinfo);
	/* get present status; this is pointless since the file to parse will not
	   be present when battery is not present.
	*/
	if (!readIntegerFrom((char *)&battinfo->statepath, "present", &battinfo->present))
		battinfo->present = 0;
}

static void parse_battinfo(BattInfo *battinfo) {
	/* get last full capacity in uWh; return in mWh */
	if (!readIntegerFrom((char *)&battinfo->statepath, "energy_full", &battinfo->fullcap))
		/* do not set this to 0 because charge percentage is calculated by
		   division by full capacity; anyway, this shoud not happen */
		battinfo->fullcap = 1;
	battinfo->fullcap /= 1000;
}

static void acquire_battinfo(ACPInfo *acpinfo) {
	BattInfo *batt_info = acpinfo->batteries;

	while (batt_info) {
		update_battstate(batt_info);
		batt_info = batt_info->next;
	}
}

static BattInfo *init_battery(char *battDir) {
	BattInfo *batt_info=malloc(sizeof(BattInfo));
		/* just crash on failed malloc... */
	
	strcpy(batt_info->statepath, battDir);
	parse_battinfo(batt_info);
	update_battstate(batt_info);
	return batt_info;
}

char * deref_sysfs_link(char * topDir, char * linkedDir) {
	DIR *dir;

	if ((dir=opendir(topDir))) {
		struct dirent *entry;
		while ((entry=readdir(dir))) {
			if (!strstr(linkedDir, entry->d_name)) {
				/* not the wanted dir */
				continue;
			}
			int actLen;
			char full_link_dirname[PATHLEN];
			char expanded_dirname[PATHLEN];
			static char full_expanded_dirname[PATHLEN];
			/* put together complete link path */
			snprintf(full_link_dirname, PATHLEN, "%s/%s", topDir, entry->d_name);
			/* deref the link */
			actLen = readlink(full_link_dirname, expanded_dirname, PATHLEN);
			expanded_dirname[actLen] = '\0';
			/* put together complete dereferenced path */
			snprintf(full_expanded_dirname, PATHLEN, "%s/%s", topDir,
				expanded_dirname);
			actLen += strlen(topDir);
			if (actLen>=PATHLEN) {
				break;
			} else if (!exists(full_expanded_dirname)) {
				break;
			}
			return full_expanded_dirname;
		}
	}
	return NULL;
}

void collect_batteries(ACPInfo *acpinfo) {
	DIR *dir;
	BattInfo **batt_chaining = &(acpinfo->batteries);
	*batt_chaining = NULL;

	if ((dir=opendir("/sys/class/power_supply"))) {
		struct dirent *entry;
		while ((entry=readdir(dir))) {
			if (!strstr(entry->d_name, "BAT")) { 
				/* not a battry dir */
				continue;
			}
			static char * full_expanded_dirname;
			full_expanded_dirname = deref_sysfs_link("/sys/class/power_supply", entry->d_name);

			*batt_chaining = init_battery(full_expanded_dirname);
			batt_chaining = &((*batt_chaining)->next);
			*batt_chaining = NULL;
		}
	}
}

void init_acpinfo(ACPInfo *acpinfo) {
	acpinfo->speed = -1;
	acpinfo->cooling=0;
	acpinfo->state=0;
	acpinfo->batt=0;
	strcpy(acpinfo->ac_path, deref_sysfs_link("/sys/class/power_supply",
		"AC0"));
	strcpy(acpinfo->thermal_path, deref_sysfs_link("/sys/class/thermal",
		"thermal_zone0"));
	acpinfo->load=0;
	acpinfo->temperature = -1;
	acpinfo->critical_temp = -1;
	acpinfo->passive_temp = -1;
	acpinfo->charge_hist = history_new(7);
		/* crash on failed malloc */
	/* The following two values should be properly initialized;
	   read all trip_point_[number]_type files and assign critical/
	   passive values thereafter.  For now, it is like this:
	   /sys/class/thermal/thermal_zone0/trip_point_0_temp (critical),
	   /sys/class/thermal/thermal_zone0/trip_point_1_temp (passive).
	   For now, however, these values are used nowhere. */
	if (readIntegerFrom(acpinfo->thermal_path, "trip_point_0_temp", &acpinfo->critical_temp)) {
		/* temperature is parsed in milli °C */
		acpinfo->critical_temp /= 1000;
	} else {
		acpinfo->critical_temp = -1;
	}
	if (readIntegerFrom(acpinfo->thermal_path, "trip_point_1_temp", &acpinfo->passive_temp)) {
		/* temperature is parsed in milli °C */
		acpinfo->passive_temp /= 1000;
	} else {
		acpinfo->passive_temp = -1;
	}
	collect_batteries(acpinfo);
}

void acquire_info(ACPInfo *acpinfo) {
    FILE *fd;
    static time_t last_info_stamp = 0;
	time_t new_time = time(NULL);

	/* see if we've been sleeping for a long time and reset history if so. */
	if (new_time-last_info_stamp>100) {
		history_clear(acpinfo->charge_hist);
	}
	last_info_stamp = new_time;

	if (readIntegerFrom(acpinfo->thermal_path, "temp", &acpinfo->temperature)) {
		/* temperature is parsed in millidegrees Celsius */
		acpinfo->temperature /= 1000;
	} else {
		acpinfo->temperature = -1;
	}
	fd = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq", "r");
	if (fd) {
		fscanf(fd, "%d", &(acpinfo->speed));
		fclose(fd);
		acpinfo->speed /= 1000;
	}
	if(!readIntegerFrom(acpinfo->ac_path, "online", &acpinfo->batt))
		acpinfo->batt = -1;
	acpinfo->load=0;
	// dump_acpinfo(acpinfo);
}

void guesstimate_remaining_time(ACPInfo *acpinfo) {
	int remaining_seconds = -1;
	double rate;
	
	rate = history_diffaverage(acpinfo->charge_hist);
	acpinfo->current_rate = rate;
	if ((acpinfo->cum_charge==2 && acpinfo->batt==1) || fabs(rate)<1e-7 ||
			isnan(rate)) {
		remaining_seconds = 0;
	} else if (acpinfo->cum_charge==0) {  /* charging */
		remaining_seconds = (int)((acpinfo->cum_fullcap-acpinfo->cum_remaining
				)/rate);
	} else if (acpinfo->cum_charge!=-1) { /* discharging */
		remaining_seconds = (int)(acpinfo->cum_remaining/rate);
	}
	
	if (remaining_seconds!=-1) {
		remaining_seconds = remaining_seconds<0?
			-remaining_seconds:remaining_seconds;
		acpinfo->remaining_minutes = (remaining_seconds/60)%60;
		acpinfo->remaining_hours = (remaining_seconds/60)/60;
	} else {
		acpinfo->remaining_minutes = acpinfo->remaining_hours = -1;
	}		
}

void compute_battStatus(ACPInfo *acpinfo) {
	BattInfo *batt_info = acpinfo->batteries;
	int new_charge = 3;
	acquire_battinfo(acpinfo);

	acpinfo->cum_rate = 0;
	acpinfo->cum_remaining = 0;
	acpinfo->cum_fullcap = 0;
	while (batt_info) {
		acpinfo->cum_rate += batt_info->currrate;
		acpinfo->cum_remaining += batt_info->currcap;
		acpinfo->cum_fullcap += batt_info->fullcap;
		new_charge = MIN(new_charge, batt_info->charge);
		batt_info = batt_info->next;
	}

	if (acpinfo->cum_remaining!=history_curval(acpinfo->charge_hist)) {
		history_addval(acpinfo->charge_hist, acpinfo->cum_remaining);
	}
	if (new_charge!=acpinfo->cum_charge) {
		/* History becomes meaningless if charge state changes */
		history_clear(acpinfo->charge_hist);
	}
	acpinfo->cum_charge = new_charge;
	acpinfo->battery_percent = (acpinfo->cum_remaining*100
		)/acpinfo->cum_fullcap;
	guesstimate_remaining_time(acpinfo);
}

#endif /* ACPI */

/* vi:sw=4:ts=4: 
*/
